	
	
	
/* 
	FEM mesh extraction for regular, cylindrical bodies.
	Compatible with Tochnog input file, see http://tochnog.sourceforge.net/
*/


/*		MODULE FOR CRYSTAL IMAGE 0.9.x BETA
		This program is copyrighted material.
		(C) 2002-2011 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>



#include "../src/defs.h"
#include "../src/dlgbox.h"
#include "../src/functions.h"
#include "../src/functions2.h"
#include "../src/findclust.h"
#include "../src/interface.h"
#include "../src/dso.h"
#include "../src/macro.h"
#include "../src/gtkhelpers.h"
#include "../src/special1.h"
#include "../src/imgbuf.h"
#include "../src/application.h"
#include "../src/secondwnd.h"
#include "../src/tiffstuff.h"




typedef struct
{
	float x; float y; float z;		/* Coordinate of the node */
	int slice;						/* Slice where the node originates from */
	float rayavg;					/* Average intensity value along the ray */
	unsigned char nodeflags;		/* 1 is interior node, 2 is upper/lower boundary node */
}
nodetype;




/* 

	A module to extract a mesh from a segmented object.
	At this time, it works on cylindrical objects. 
	Extensions for convex objects may be added at a later time.

	HOW IT WORKS

	1) PASS 1 - find the center line
		To find the center line, the centroid of the segmented object is found
		in each slice. We therefore get xc(z) and yc(z). Now, we perform a lsq fit
		into both xc:z and yc:z and get two functions that define the center point
		for each slice. This way, we can eliminate a moderate tilt.

	2) PASS 2 - extract the quads
		In each slice separately, we emit probing "rays" from the center point
		towards the image margin. When the image intensity of the probing ray
		increases above the thresahold, we found one node. Continuing further outward,
		when the intensity falle below the threshold, we found another node.
		Designate these nodes ar distances from centroid r1 and R1.
		Now, we start another ray, offset at angle theta and obtain r2 and R2
		the same way. Once we have done the same one slice above, we have completed
		one quad, defined by 8 nodes.

	3) PASS 3 - Manage and output quads, nodes, and material information


*/


/*	The module can be compiled with the following entry in make file. Just type make meshextract.dso

# Handled out of the ordinary. This line compiles the dummy module (dso sample function.
# Once it is compiled, manually copy the dso into your dso directory, which is
# either /usr/local/cimage/dso or ~/dso

meshextract.dso: meshextract.c
	$(CC) $(GTK_CFLAGS) $(DEBUGFLG) -c -fPIC meshextract.c
	$(CCLD) $(imager_LDADD) -shared -fPIC -o meshextract.dso meshextract.o
	chmod a-x meshextract.dso

*/




void Spline_new (spline_block *spl, int numpts, float *v)
{
int i,m;
float *gamma;
float *delta;
float *bigd;

	m = numpts-1;
	spl->npoints = numpts-1;
	spl->coeff_a = calloc (numpts+2, sizeof(float));
	spl->coeff_b = calloc (numpts+2, sizeof(float));
	spl->coeff_c = calloc (numpts+2, sizeof(float));
	spl->coeff_d = calloc (numpts+2, sizeof(float));
	gamma = calloc (numpts+2, sizeof(float));
	delta = calloc (numpts+2, sizeof(float));
	bigd = calloc (numpts+2, sizeof(float));

	/* Calculate the gamma coefficients */

	gamma[0] = 0.5;									/* left endpoint */
	for (i=1; i<m; i++)
		gamma[i] = 1.0/(4.0 - gamma[i-1]);

	gamma[m] = 1.0/(2.0 - gamma[m-1]);				/* right endpoint */


	/* Recursively calculate the delta vector; builds on gamma. */

	delta[0] = 3.0*gamma[0]*(v[1]-v[0]);
	for (i=1; i<m; i++)
		delta[i] = gamma[i]*( 3.0*(v[i+1]-v[i-1]) - delta[i-1]);

	delta[m] = gamma[m]*( 3.0*(v[m]-v[m-1]) - delta[m-1]);

	/* Recusively calculate the diagonal elements of the D matrix,
		these are already the coefficients b. */

	bigd[m] = delta[m];
	for (i=m-1; i>=0; i--)
		bigd[i] = delta[i] - gamma[i]*bigd[i+1];

	/* Last step: Create the coefficients a,b,c,d from D */

	for (i=0; i<m; i++)			/* We don't carry the last vertex, see also the spl->npoints assigment */
	{
		spl->coeff_a[i] = v[i];			/* Continuity: coincide with vertex when t=0 */
		spl->coeff_b[i] = bigd[i];			/* Extra step, but increases readability */
		spl->coeff_c[i] = 3.0*(v[i+1]-v[i]) - 2.0*bigd[i] - bigd[i+1];
		spl->coeff_d[i] = 2.0*(v[i]-v[i+1]) + bigd[i] + bigd[i+1];

	}

	/* Free the intermediate arrays */

	free (gamma); free (delta); free (bigd);

}


float Spline_get (spline_block *spl, int segment, float t)
{
float a,b,c,d, r;

	if (segment > spl->npoints)
	{
		dp (0, "Cubic spline: Attempt to access invalid segment %d\n", segment);
		return 0.0;
	}
	if ((t<0.0) || (t>1.0))
	{
		dp (0, "Cubic spline: Attempt to access invalid fraction %f of segment %d\n", t, segment);
		return 0.0;
	}

	a = spl->coeff_a[segment];			/* This is to improve readability */
	b = spl->coeff_b[segment];
	c = spl->coeff_c[segment];
	d = spl->coeff_d[segment];

	r = a + t*(b + t*(c + t*d));		/* Interpolation function */
	return r;

}


void Spline_free (spline_block *spl)
{

	spl->npoints = 0;
	if (spl->coeff_a) free (spl->coeff_a);
	spl->coeff_a = NULL;
	if (spl->coeff_b) free (spl->coeff_b);
	spl->coeff_b = NULL;
	if (spl->coeff_c) free (spl->coeff_c);
	spl->coeff_c = NULL;
	if (spl->coeff_d) free (spl->coeff_d);
	spl->coeff_d = NULL;

}





/*          ###     ###    ####    #####           #####   #   #   #   #    ###    #####    ###     ###    #   #          */
/*         #   #   #   #   #   #   #               #       #   #   #   #   #   #     #       #     #   #   #   #          */
/*         #       #   #   #   #   #               #       #   #   ##  #   #         #       #     #   #   ##  #          */
/*         #       #   #   ####    ####            ####    #   #   # # #   #         #       #     #   #   # # #          */
/*         #       #   #   # #     #               #       #   #   #  ##   #         #       #     #   #   #  ##          */
/*         #   #   #   #   #  #    #               #       #   #   #   #   #   #     #       #     #   #   #   #          */
/*          ###     ###    #   #   #####           #        ###    #   #    ###      #      ###     ###    #   #          */
/*                                                                                                                        */


int *meshfx, *meshfy;		/* Centroids for mesh generation */
unsigned char *meshflg;	/* Whether a centroid was found */
image_type meshx_labelimg;		/* Image where we can see the intersections */

int isign (int i) { if (i>0) return 1; else if (i<0) return -1; else return 0; }


/* node_equal compares two nodes n1 and n2 and returns 1 if they are at the same spatial location */


int node_equal (nodetype n1, nodetype n2)
{
double E;			/* Euclidean distance, squared */

	E = SQR(n1.x-n2.x) + SQR(n1.y-n2.y) + SQR(n1.z-n2.z);
	if ( E < 1e-16 )
		return 1;
	else
		return 0;
}


/* nlookup tries to find node "node" in node list nlist (with nmax entries),
	nlookup returns -1 if the node does not exist and a number from 0 to nmax if 
	the node has been found. The returned value is the index in nlist */

int nlookup (nodetype node, nodetype *nlist, int nmax)
{
int i;

	if (nmax==0) return -1;
	i=0;
	while ((i<nmax) && (!node_equal (node, nlist[i]))) i++;

	if ((i<nmax) && (node_equal (node, nlist[i]) )) return i;
	return -1;
}


/* Prepare the group array: Get all image values above the threshold
	then split the min/max we found into groupcnt discrete steps */

void prep_groups (image_type *img, float thresh, int groupcnt, float *grp)		/* Prepare the group aray */
{
float min, max, buf;
int x,y,z, first;

	first=1;
	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if ( (buf=readbuf_flt (img,x,y,z)) >= thresh)
				{
					if (first)
					{
						min=max=buf; first=0;
					}
					else
					{
						min = MIN (min,buf);
						max = MAX (max,buf);
					}
				}
			}

	/* Now compute the group interval start values */

	grp[0]=0;		/* Group 0 always catches the values that are too low */

	for (x=0; x<=groupcnt; x++)
	{
		grp[x+1]=min + x*(max-min)/groupcnt;
	}
}


/* Return the group number of which val is a member */

int member_of_group (float *grp, float val, int groupcnt)
{
int i;

	i=0;
	while ((i<groupcnt) && (val>grp[i])) i++;
	return i;
}


/* Compute the image average inside a 2D bounding box. Consider only values above the threshold */

float get_image_average (image_type *img, float thresh, int x1, int y1, int x2, int y2, int z)
{
int x,y;
float buf, avg; int cnt;

	avg=0.0; cnt=0;
	for (y=MIN(y1,y2); y<=MAX(y1,y2); y++)
		for (x=MIN(x1,x2); x<=MAX(x1,x2); x++)
		{
			buf = readbuf_flt (img, x,y,z);
			if (buf >= thresh)
			{
				avg += buf; cnt++;
			}
		}

	if (cnt==0)
		return 0;
	else
		return avg/cnt;
}




/* Run along a line from (x0,y0,z) to (x1,y1,z) using Bresenham algorithm.
	Fill arrays xb and yb with the coordinates and return number of pixels */

int create_bresenham_line (image_type *img, int x0, int y0, int x1, int y1, int z, int *xb, int *yb)
{
int i,x,y,len;
int dx,ax, dy, ay,sgx,sgy,d;


	i=0;
	dx=x1-x0; dy=y1-y0; ax=2*abs(dx); ay=2*abs(dy);
	sgx=isign(dx); sgy=isign(dy); x=x0; y=y0;
	if (ax>ay)        /* x dominant */
	{
		d = ay - ax/2; len = abs(dx);
		while ((x!=x1) && (x>1) && (x<img->xmax-1))
		{
			xb[i]=x; yb[i]=y; i++;
			if (d>=0) { y+=sgy; d-=ax; }
			x+=sgx; d+=ay;
		}
	}
	else              /* y dominant */
	{
		d = ax - ay/2; len = abs(dy);
		while ((y!=y1) && (y>1) && (y<img->ymax-1))
		{
			xb[i]=x; yb[i]=y; i++;
			if (d>=0) { x+=sgx; d-=ay; }
			y+=sgy; d+=ax;
		}
	}

	return i;
}


/* Follow_ray follows a Bresenham line from xc,yc to x1,y1 in slice z. The bresenham line is
	created inside this function and temporarily stored in xb,yb; the line is nsteps pixels long.
	Thresh is the threshold value at which the inner and outer wall is detected.
	Aniso is the Z anisotropy and is used as multiplier for the integer z
	qn is the node storage, a nodetype array with 8 elements to store the result(i.e.,
		the two edges found in each run of follow_ray and the average image value).
	targ1 and targ2 are the node numbers and must be between 0 and 7.

	In follow_ray, we walk nsteps along (xb,yb,z) and check for the threshold.
		At the intersection with the edge, we record a node. First, the ray crosses
		a transition from low background to high object (inner edge, at index targ1), 
		which we record as inner node. Then, the ray crosses the transtion back to 
		background which we record as outer wall node (at index targ2).
		The states are as follows:
		state==0:	beginning ray, all pixels up to here stayed below threshold
		state==1:	Threshold exceeded. We are now inside the walls (average density is being recorded)
		state==2:	Dropped below threshold. Outside the walls, node has been recorded, finish run.

	If interpolation is requested, we calculate a weighted average of the actual pixel
		with the previous pixel by linear interpolation (new August 2011)

	TODO: Follow_ray should return an error if no intersection with the wall was found.

*/


int follow_ray (image_type *img, float thresh, float xc, float yc, float x1, float y1, int z, int mode, nodetype *qn, int targ1, int targ2)
{
int xb[2*img->xmax+1], yb[2*img->ymax+1];
int state;
float i0, i1, avg, alpha, beta;
float buf;
float fx, fy, dx, dy;						/* Non-integer step coords and step increments */
int cnt, i, nsteps, ix, iy;
rgbtriplet rgb;


		if (mode==0)			/* Non-interpolating: Use Bresenham */
			nsteps = create_bresenham_line (img, IROUND(xc), IROUND(yc), IROUND(x1), IROUND(y1), z, xb, yb);
		else
		{						/* Interpolating, use unit-step floating-point coordinates */

			dx = x1 - xc; dy = y1 - yc;
			nsteps = IROUND(sqrt(ARITHM(dx,dy)));			/* Length of line (nearest integer) */
			dx /= nsteps; dy /= nsteps;
			fx = xc; fy = yc;
		}

		dp (3, "Slice %d: Ray from (%d,%d) to (%d,%d)\n",z,xc,yc,x1,y1);

		state=0 ; avg=0.0; cnt=0;
		for (i=0; i<nsteps; i++)
		{

			/* Depending on mode (interpolating or not), get the image value at the current location */

			if (mode==0)
				buf = readbuf_flt(img,xb[i],yb[i],z);		/* Integer read from Bresenham line */
			else
				buf = ireadbuf (img, fx,fy,z, intp_pref);	/* Non-integer read from fractional line */

			if  ((state==0) && (buf>thresh))				/* Threshold exceeded for the first time -> rising edge */
			{

				if (mode==0)			/* Store as-is, no interpolation */
				{
					qn[targ1].x = xb[i];	qn[targ1].y = yb[i];  /* Quad node designation for coordinates @ rising edge */
				}
				else					/* Interpolate between pixels */
				{
					i0 = ireadbuf (img,fx-dx,fy-dy,z, intp_pref); 						/* Pixel before state change */
					i1 = buf; 															/* Pixel after state change */
					beta = (i1-thresh)/(i1-i0);											/* Interpolation value */
					if ((beta>1.0) || (beta<0.0)) beta=1.0;								/* Should never happen */
					alpha = 1.0-beta;
 					qn[targ1].x = fx - beta*dx;
 					qn[targ1].y = fy - beta*dy;
					if (mode==2)
					{
						ix = IROUND(fx-dx); iy = IROUND(fy-dy);
						rgb = (rgbtriplet){0, (int)(255*beta), 0};
						writebuf (&meshx_labelimg, &rgb, ix,iy, z);
						ix = IROUND(fx); iy = IROUND(fy);
						rgb = (rgbtriplet){0, (int)(255*alpha), 0};
						writebuf (&meshx_labelimg, &rgb, ix,iy, z);
					}
				}
				qn[targ1].z=z;			/* The z coordinate is never interpolated */
				qn[targ1].slice=z;		/* independently store the slice */
				state=1;
			}

			if (state==1)										/* While we are inside the object, */
			{												/* average intensity along this ray */
				avg += buf; cnt++;
			}

			if ((state==1) && (buf<thresh))
			{
				if (mode==0)			/* Store as-is, no interpolation */
				{
					qn[targ2].x = xb[i];	qn[targ2].y = yb[i];  /* Quad node designation for coordinates @ falling edge */
				}
				else					/* Interpolate between pixels */
				{
					i0 = ireadbuf (img,fx-dx,fy-dy,z, intp_pref); 						/* Pixel before state change */
					i1 = buf; 															/* Pixel after state change */

					alpha = (i0-thresh)/(i0-i1);														/* Interpolation value */
					if ((alpha>1.0) || (alpha<0.0)) alpha=1.0;											/* Should never happen */
					beta = 1.0-alpha;
 					qn[targ2].x = fx - beta*dx;
 					qn[targ2].y = fy - beta*dy;
					if (mode==2)
					{
						ix = IROUND(fx-dx); iy = IROUND(fy-dy);
						rgb = (rgbtriplet){0, (int)(255*beta), 0};
						writebuf (&meshx_labelimg, &rgb, ix,iy, z);
						ix = IROUND(fx); iy = IROUND(fy);
						rgb = (rgbtriplet){0, (int)(255*alpha), 0};
						writebuf (&meshx_labelimg, &rgb, ix,iy, z);
					}
				}
				qn[targ2].z=z;			/* The z coordinate is never interpolated */
				qn[targ2].slice=z;		/* independently store the slice */
				state=2;
			}

			if (mode>0)			/* Update line coords in interpolating mode */
			{
				fx += dx; fy += dy;
			}

		}
		if (cnt>0) avg/=cnt; else avg=0;
		qn[targ1].rayavg = avg;
		qn[targ2].rayavg = avg;

		if (state<2) dp (1, "ERROR: Ray from (%d,%d) to (%d,%d) did not find 2 edges!\n",xc,yc,x1,y1);
		if (state<2) return -1;		/* If we didn't reach state 2, we did not find a complete wall */

		return 0;			/* Exit with no error */

}




/* node_interpolate creates new nodes in between raw nodes by spline interpolation.
		rawnodes contains the nides in the order inner - outer - inner - outer ...
		sn holds the interpolated nodes, same order
		sncnt has the total number of raw nodes, note that one inner node exists
				at rawnodes[sncnt] and one outer node at rawnodes[sncnt+1], those are the
				same as rawnodes[0] and rawnodes[1] and close the circle.
		numinterp is the number of new nodes to create in between two raw nodes
		the function returns the number of final nodes analogous to sncnt.
*/


int node_interpolate (nodetype *rawnodes, nodetype *sn, int sncnt, int numinterp)
{
int i,j,k;
float x,y;
int N;
float v_x[sncnt+10];				/* One dimension of vertex coordinates */
float v_y[sncnt+10];				/* The other dimension of vertex coordinates */
spline_block splx, sply;		/* Spline coefficients */


	/* if numinterp==0, we simply copy */

	if (numinterp==0)
	{
		for (i=0; i<sncnt+2; i++)
		{
			sn[i] = rawnodes[i];
			dp (3, "        %3d      %5.1f %5.1f\n", i, sn[i].x, sn[i].y);
		}

		return sncnt;
	}

	/* Here, numinterp > 0 and we need to interpolate.
		Start with the inner nodes */

	j=0;
	for (i=0; i<sncnt+2; i+=2)			/* Inner nodes at even indices */
	{
		v_x[j  ] = rawnodes[i].x;
		v_y[j++] = rawnodes[i].y;
	}

	Spline_new (&splx, j, v_x);		/* Preapre spline coeffs for x */
	Spline_new (&sply, j, v_y);		/* Preapre spline coeffs for y */



	N=0;
	for (i=0; i<j-1; i++)
	{
		for (k=0; k<=numinterp; k++)	/* k=0 is the original node, subsequent k's interpolate */
		{
			x = Spline_get (&splx, i, (float)k/(float)(numinterp+1));
			y = Spline_get (&sply, i, (float)k/(float)(numinterp+1));
			sn[N].x = x;
			sn[N].y = y;
			N+=2;						/* Advance by 2 to leave space for the interleaved outer nodes */
		}
	}

	Spline_free (&splx);
	Spline_free (&sply);

	/* Next, do the same again with the outer nodes */

	j=0;
	for (i=0; i<sncnt+2; i+=2)			/* Outer nodes at odd indices */
	{
		v_x[j  ] = rawnodes[i+1].x;
		v_y[j++] = rawnodes[i+1].y;
	}

	Spline_new (&splx, j, v_x);		/* Preapre spline coeffs for x */
	Spline_new (&sply, j, v_y);		/* Preapre spline coeffs for y */

	N=0;
	for (i=0; i<j-1; i++)
	{
		for (k=0; k<=numinterp; k++)	/* k=0 is the original node, subsequent k's interpolate */
		{
			x = Spline_get (&splx, i, (float)k/(float)(numinterp+1));
			y = Spline_get (&sply, i, (float)k/(float)(numinterp+1));
			sn[N+1].x = x;
			sn[N+1].y = y;
			N+=2;						/* Advance by 2, fill the interleaved outer nodes */
		}
	}

	Spline_free (&splx);
	Spline_free (&sply);


	/* Close the circle by copying nodes #0/#1 into nodes #N+1 and #N+2 */

	sn[N  ] = sn[0];
	sn[N+1] = sn[1];

	dp (3, " %3d    Node     x     y    \n", rawnodes[0].slice);
	for (i=0; i<N+2; i++)
	{
		dp (3, "        %3d      %5.1f %5.1f", i, sn[i].x, sn[i].y);
		if (i & 0x01)
			dp (3, "    outer\n");
		else
			dp (3, "    inner\n");
	}

	return N;


}


/* A helper function for node_interpolate:
	labels a node in the viz image when mode==2 is selected */

void label_node (nodetype qn)
{
int ix, iy, z;
rgbtriplet rgb = {0xff, 0x40, 0x40};


	z = qn.slice;
	ix = IROUND(qn.x); iy = IROUND(qn.y);
	writebuf (&meshx_labelimg, &rgb, ix,iy, z);

}




/**************************************************************************************************************************/

/* Tochnog extra data. Tochnog requires some information at the beginning of the file
	(written by tochnog_header) and some control info at the end (written by tochnog_footer)
*/

void tochnog_header (FILE *FP)
{
	fprintf (FP, "echo -yes\n");
	fprintf (FP, "number_of_space_dimensions 3\n");
	fprintf (FP, "condif_temperature\n");
	fprintf (FP, "groundflow_pressure\n");
	fprintf (FP, "materi_velocity\n");
	fprintf (FP, "materi_stress\n");
	fprintf (FP, "materi_strain_elasti\n");
	fprintf (FP, "materi_strain_total\n");
	fprintf (FP, "end_initia\n");
	fprintf (FP, "\n\n\n");
}

void tochnog_footer (FILE *FP, int index, int backend)
{

	fprintf (FP, "\n\n\n");
	switch (backend)
	{
		case 0:
					/* GiD is the default. No specific instruction needed */
			break;
		case 1:
			fprintf (FP, "control_print_vtk               %2d  -yes\n", index);
			break;
		case 2:
			fprintf (FP, "control_print_dx               %2d  -yes\n", index);
			break;
	}

	fprintf (FP, "control_timestep               %2d  0.1 1.\n", index);
	fprintf (FP, "control_timestep_iterations    %2d  5\n", index);
	fprintf (FP, "control_print                  %2d  -time_current -post_node_rhside_ratio\n", index);
	fprintf (FP, "end_data\n");
}


/* NOTE: The following material definition is created once per group.
	Here, we are using some arbitrary material properties. 
	Note that a high group number corresponds to high density. 
	Therefore, we associate a high Young's modulus with the high groups.
	For development, we use the approximate modulus of rubber, 10 - 100 MPa.

	Depending on the application (bone, soft tissue, tendon.....) there should
	be realistic material property definitions.
	If the number of groups is low, hand-editing is recommended.
*/

void tochnog_material (FILE *FP, int grp, float min_young, float max_young)
{
	fprintf (FP, "group_type %d   -materi\n",grp);
	fprintf (FP, "group_materi_density %d   %g\n",grp,0.03);
	fprintf (FP, "group_materi_elasti_young %d   %g\n",grp, min_young + (max_young-min_young)*grp);
	fprintf (FP, "group_materi_elasti_poisson %d   %g\n",grp,0.45);
	fprintf (FP, "group_materi_memory %d   -updated\n",grp);
	fprintf (FP, "group_materi_expansion_linear %d   %g\n",grp,9.0e-5);
	fprintf (FP, "\n");
}



/* STL FILE CODE -- trying to be widely compatible with Tochnog sections  */


void stl_header (FILE *FP)
{
	fprintf (FP, "solid extracted_mesh\n");
}


void stl_footer (FILE *FP, int index, int backend)
{
	fprintf (FP, "endsolid extracted_mesh\n");
}

void stl_material (FILE *FP, int grp, float min_young, float max_young)
{

	/* Nothing to do here */

}

/* The facet is less compatible as it combines "element" and "node".
	A facet is given by three vertices IN COUNTER-CLOCKWISE ORDER!
	We use the nodetype to package the vertices.
*/

void stl_facet (FILE *FP, nodetype v1, nodetype v2, nodetype v3, float scale)
{
float nx, ny, nz;								/* Normal vector */
float v1x, v1y, v1z, v2x, v2y, v2z;			/* Intermediate */
float len;


	v1x = v2.x - v1.x; v1y = v2.y - v1.y; v1z = v2.z - v1.z; 
	v2x = v3.x - v1.x; v2y = v3.y - v1.y; v2z = v3.z - v1.z; 
	nx = v1y*v2z - v1z*v2y;
	ny = v1z*v2x - v1x*v2z;
	nz = v1x*v2y - v1y*v2x;
	len = sqrt (SQR(nx)+SQR(ny)+SQR(nz));
	if (len>0)
	{
		nx /= len;
		ny /= len;
		nz /= len;
	}

	fprintf (FP, "facet normal %g %g %g\n", nx,ny,nz);
	fprintf (FP, "   outer loop\n");
	fprintf (FP, "      vertex %g %g %g\n", v1.x*scale, v1.y*scale, v1.z*scale);
	fprintf (FP, "      vertex %g %g %g\n", v2.x*scale, v2.y*scale, v2.z*scale);
	fprintf (FP, "      vertex %g %g %g\n", v3.x*scale, v3.y*scale, v3.z*scale);
	fprintf (FP, "   endloop\n");
	fprintf (FP, "endfacet\n");

}


/* Write one quad for Tochnog. Unlike STL, this one uses node (=vertex) numbers */

void tochnog_element (FILE *FP, int elem, int *enodes)
{
int k;

	fprintf (FP, "element %d    -hex8    ",elem);
	for (k=0; k<8; k++) fprintf (FP, "  %d",enodes[k]);
	fprintf (FP, "\n");

}


/* The same for STL. Here, we use the quad_nodes (qn) instead of the node numbers */

void stl_element (FILE *FP, int elem, nodetype *qn, float scale)
{

	stl_facet (FP, qn[0], qn[4], qn[1], scale);		/* Inward-facing triangle 1 */
	stl_facet (FP, qn[1], qn[4], qn[5], scale);		/* Inward-facing triangle 2 */

	stl_facet (FP, qn[2], qn[3], qn[6], scale);		/* Outward-facing triangle 1 */
	stl_facet (FP, qn[3], qn[7], qn[6], scale);		/* Outward-facing triangle 2 */

}


/* For STL, we create a top and bottom plate. Note that we have different normal directions
   for top and bottom plates. */

void stl_plate (FILE *FP, nodetype *sn, float scale, int nnodes, int is_bottom)
{
int ray;

	for (ray=0; ray<nnodes; ray++)	
	{
		if (is_bottom)
		{
			stl_facet (FP, sn[2*ray  ], sn[2*ray+2], sn[2*ray+1], scale);
			stl_facet (FP, sn[2*ray+1], sn[2*ray+2], sn[2*ray+3], scale);
		}
		else
		{
			stl_facet (FP, sn[2*ray  ], sn[2*ray+1], sn[2*ray+2], scale);
			stl_facet (FP, sn[2*ray+1], sn[2*ray+3], sn[2*ray+2], scale);
		}

	}

}




/**************************************************************************************************************************/


/* This is the code for PASS 1, to extract the centroid line */

void meshx_pass1 (image_type *img, float thresh, int mode)
{
int x,y,z,z1;
float xc, yc, n;
float buf;
fitresult fit;
double xdata[img->zmax], ydata[img->zmax];
rgbtriplet rgb;


	meshfx = calloc (img->zmax, sizeof (int));
	meshfy = calloc (img->zmax, sizeof (int));
	meshflg = calloc (img->zmax, sizeof (unsigned char));

	rgb = (rgbtriplet){255,255,255};		/* Pure white for labeling */

	for (z=0; z<img->zmax; z++)				/* For each slice... */
	{

		xc=0; yc=0; n=0;
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,z);		/* Read image value */
				if (buf >= thresh)					/* Is a feature pixel */
				{
					xc += x; yc+= y; n++;
					if (mode==2)
						writebuf (&meshx_labelimg, &rgb, x, y, z);		/* Every pixel above threshold is white */
				}
			}

		/* Sum computation finished. Store in array */

		if (n>0)
		{
			xc /= n; yc /= n;
			meshfx[z] = xc;
			meshfy[z] = yc;
			meshflg[z] = 1;
		}

	}

	/* Now we do a least squares fit to determine the functions xc and yc of z */

	z1=0;
	for (z=0; z<img->zmax; z++)
		if (meshflg[z])
		{
			xdata[z1]=z;
			ydata[z1]=meshfx[z];
			z1++;
		}
	fit = lsqfit (xdata, ydata, 0, z1-1);
	for (z=0; z<img->zmax; z++)
	{
		meshfx[z] = (int) (0.4999+z*fit.H+fit.intcpt);
	}
	dp (1, "Centroid functions: %d slices. X-intercept %f, x-slope %f, ",z1,fit.intcpt,fit.H);

	z1=0;
	for (z=0; z<img->zmax; z++)
		if (meshflg[z])
		{
			xdata[z1]=z;
			ydata[z1]=meshfy[z];
			z1++;
		}
	fit = lsqfit (xdata, ydata, 0, z1-1);
	for (z=0; z<img->zmax; z++)
	{
		meshfy[z] = (int)(0.4999+z*fit.H+fit.intcpt);
	}
	dp (1, "y-intercept %f, y-slope %f\n",fit.intcpt,fit.H);


}


/* This is the code for PASS 2, the main section to extract the nodes and quads  */
/* The threshold is the threshold at which a transition from background to object
		is identified. Numrays are the number of radial rays and determine the
		radial discretization. zskip allows to skip Z slices, and a higher value
		gives a lower axial discretization. press is the pressure and enables to apply nodal pressure.
		The filename is the file to which to write the Tochnog input file.

	Follow the rays. Each quad is delineated by 4 rays and has 8 nodes.
		The first ray is at angle theta and the current slice. It provides nodes 0 and 2.
		The second ray is at angle theta+dtheta and the current slice. It provides nodes 1 and 3.
		The third ray is at angle theta and the next slice up.  It provides nodes 4 and 6.
		The fourth ray is at angle theta+dtheta and the next slice up.  It provides nodes 5 and 7.

		A sketch of the node order follows (rays run from left to right).

		 lower z: -----0-----2--->			upper z: -----4-----6--->
						|     |   			              |     |
				   -----1-----3--->			     -----5-----7--->

		Therefore, the inner wall nodes are 0,1,4,5. The lower nodes are 0,1,2,3 and the
		upper nodes are 4,5,6,7. This is important when setting the nodeflags.

		To avoid confused nodes, we need to run along one slice exactly ONCE
		and store the nodes; the quads are then written when the next slice has been completed,
		and the just-completed slice scrolls down.

		TODO: We need to identify rays that don't intersect with the object (hole!)
		and disable the corresponding quad(s) */



void meshx_pass2 (image_type *img, float thresh, int numrays, int numinterp, int zskip, int groupcnt, float xysize, float press, float min_young, float max_young, float aniso, const char *filename, char backend, char mode)
{
int z,lastz,i, ray, m, prevz;
int nodemax;							/* Counts how many nodes we have in the list */
float xc, yc, x1, y1;					/* Start and end of each one ray */
int k,ncnt,n,elem, quadgrp, nnodes;
float theta, dtheta;
float xofs[numrays+10], yofs[numrays+10];
float quadavg;
nodetype *nodelist;
nodetype qn[8];							/* Intermediate node processing */
nodetype rawnodes[2*numrays+10];		/* Slice nodes (raw) without interpolation */
nodetype *sn;							/* Slice nodes, after interpolation, for processing */
nodetype *psn;							/* Slice nodes, previous slice (copied from sn[]) */
int sncnt;								/* Counter for raw nodes */
int snodecnt;							/* Counter for all (raw+interpolated) nodes */
int enodes [8];
int *elemnum, *elemgrp;  /* Element numbers  */
float *groupint;		/* Lower intensity bound of a group */
int	*groupocc;			/* Group occupancy */
FILE *FP;



	if (zskip<1) zskip=1;			/* Force Z increment of at least 1 */


	nnodes = numrays*(1+numinterp);		/* The number of final nodes we'll have per slice */

	for (k=0; k<8; k++) qn[k].nodeflags=0;
	
	elemnum=calloc (nnodes*img->zmax+100, sizeof (int));
	elemgrp=calloc (nnodes*img->zmax+100, sizeof (int));
	sn  = calloc (2*nnodes+10, sizeof (nodetype));
	psn = calloc (2*nnodes+10, sizeof (nodetype));
	groupint = calloc (groupcnt+100, sizeof (float));
	groupocc = calloc (groupcnt+100, sizeof (int));

	/* Although the allocation size is constant, dynamically allocate the 
		node list so that we can check for errors. Also, we can make a conservative
		estimate as for the number of nodes we need, which is guaranteed to be
		less than 2 nodes per ray (plus the interpolated ones). */

	nodemax = ( 2*(nnodes+5)*(img->zmax+10) );		/* a few extra won't hurt */
	nodelist = calloc (nodemax+10, sizeof(nodetype));
	if (!nodelist) fatal_memerr();
	
	prep_groups (img, thresh, groupcnt, groupint);	/* Prepare the group array */

	for (k=0; k<8; k++)
	{
		qn[k].x=0; qn[k].y=0; qn[k].z=0; qn[k].slice=0;
		qn[k].rayavg=0; qn[k].nodeflags=0;
	}

	dtheta = 360.0/numrays;		/* angular increment between rays */
	m = MIN(img->xmax/2, img->ymax/2)-3;		/* Max length of ray, stay away from boundary */

	theta=0;
	for (ray=0; ray<numrays; ray++)			/* Pre-compute the ray offsets */
	{
		xofs[ray] = m*COS(theta);
		yofs[ray] = m*SIN(theta);
		theta += dtheta;
	}

	for (z=0; z<img->zmax; z+=zskip)		/* Simulate run over each slice to determine last slice */
	{
		if (meshflg[z]) lastz=z;
	}


	FP = fopen (filename, "wt");
	if (backend==3)
		stl_header (FP);
	else
		tochnog_header (FP);

	ncnt=0;  		/*Node counter */
	elem=0;  		/* Element counter */
	prevz=-1;		/* Indicate that we haven't processed a full slice yet */


	for (z=0; z<img->zmax; z+=zskip)		/* Run over each slice where we found a centroid */
	{

		if (meshflg[z])
		{
			xc = meshfx[z]; yc = meshfy[z];		/* The centroid */


			/* Run over one full slice, send out probing rays, store the slice nodes
				in the order as found: inner - outer - inner - outer ...
				Note that the last node pair repeats the firts ray at theta=0
				The odd storage in qn is a leftover from a previous approach 
			*/

			sncnt=0;
			for (ray=0; ray<numrays; ray++)
			{
				x1 = xc + xofs[ray]; y1 = yc + yofs[ray];						/* from (xc,yc) to (x1,y1) in slice z */
				follow_ray (img, thresh, xc,yc, x1,y1, z, mode, qn, 0, 1);		/* one ray */
				rawnodes[sncnt] = qn[0];										/* Inner node */
				rawnodes[sncnt+1] = qn[1];										/* Outer node */
				rawnodes[sncnt].z *= aniso;									/* Factor in the anisotropy */
				rawnodes[sncnt+1].z *= aniso;

				//dp (2, "Ray from %5.1f,%5.1f to %5.1f,%5.1f in slice %d: Nodes at (%5.1f,%5.1f) and (%5.1f,%5.1f)\n", xc,yc,x1,y1,z,qn[0].x,qn[0].y,qn[1].x,qn[1].y);

				sncnt+=2;
			}

			/* Now, copy the first node pair into last+2 to force a closed circle
					DON'T TOUCH SNCNT AFTER THIS POINT!   */

			rawnodes[sncnt] = rawnodes[0];
			rawnodes[sncnt+1] = rawnodes[1];


			/* Done collecting the nodes for one slice. Now, we create the sn node list from rawnodes
				-- either by copying rawnodes into sn (numinterp==0)
				-- or by doing two spline interpolations (inner and outer ring) (numinterp>0)
			*/

			snodecnt = node_interpolate (rawnodes, sn, sncnt, numinterp);
			dp (2, "Interpolated in slice %d: Got %d interpolated nodes from %d raw nodes\n",z,snodecnt,sncnt);


			/* At this point, we have a full interpolated set of nodes. First,
				handle the node flags to mark inner nodes (0x01) and boundary nodes (0x02).
				Also, fill in the z coordinate, since all nodes are planar. */

			for (ray=0; ray < snodecnt+2; ray++)
			{
				sn[ray].slice=z;
				sn[ray].z = z*aniso;

				if (ray & 0x01)
					sn[ray].nodeflags = 0;		/* Outer node, just reset nodeflags */
				else
					sn[ray].nodeflags = 1;			/* Inner node, resets nodeflags, too */

				if ((z==0) || (z==lastz))				/* boundary node */
					sn[ray].nodeflags |= 2;

				if (mode==2) label_node (sn[ray]);
			}


			if ((backend==3) && (z==0))			/* Create a bottom mesh of facets */
			{
				stl_plate (FP, sn, xysize*1e-3, nnodes, 1);
			}


			/* Once we have a minimum of 2 slices, we start writing elements and doing node management */
			/* The main job here is to arrange the nodes in the specific Tochnog quad order,
				see comments above this function. Although this is apparent unnecessary work, we use
				the qn array with 8 members to sort the sn nodes. Nice and clean: qn[0] to qn[7]
				create one element. */

			if (z>0)
			{

				for (ray=0; ray<nnodes; ray++)		/* "ray" is a misnomer, because it's really the index into the sn array */
				{
					qn[0] = psn[2*ray];		qn[2] = psn[2*ray+1];					/* Ray 1 (lower slice) */
					qn[1] = psn[2*ray+2];		qn[3] = psn[2*ray+3];					/* Ray 2 (lower slice) */
					qn[4] =  sn[2*ray];		qn[6] =  sn[2*ray+1];					/* Ray 3 (upper slice) */
					qn[5] =  sn[2*ray+2];		qn[7] =  sn[2*ray+3];					/* Ray 4 (upper slice) */

					/* Get average image values inside the 2D bounding box in both slices,
						then average the two slices. */

					quadavg = get_image_average (img, thresh, IROUND(qn[0].x), IROUND(qn[0].y), IROUND(qn[3].x), IROUND(qn[3].y), IROUND(qn[0].z) );
					quadavg += get_image_average (img, thresh, IROUND(qn[4].x), IROUND(qn[4].y), IROUND(qn[7].x), IROUND(qn[7].y), IROUND(qn[4].z) );

					quadavg /=2;

					/* Quadavg now contains the average intensity inside the quad. 
						Now determine to which material group this value belongs */

					quadgrp = member_of_group (groupint, quadavg, groupcnt);
					dp (2, "Element %d has average intensity of %f and group %d\n",elem, quadavg,quadgrp);
					groupocc[quadgrp]++;			/* Count how many members this group has */


					/* Node management: Look up the eight nodes with nlookup.
					If nlookup returns a positive number, we already have
					the node in the list and simply record its number for the 
					element definition in enodes. If nlookup returns a -1,
					we have a new node. We need to enter the node into 
					the nodelist and specify if it is an inner node (for pressure) */

					for (k=0; k<8; k++)
					{
						n = nlookup (qn[k], nodelist, ncnt);
						if (n<0)						/* New node */
						{
							nodelist[ncnt] = qn[k];
							enodes[k]=ncnt;
							ncnt++;
							if (ncnt>nodemax)  fatal_memerr ();
						}
						else
							enodes[k]=n;
					}

					/* Write the element to file. Note that the node order matters! */

					if (backend==3)
						stl_element (FP, elem, qn, xysize*1e-3);
					else
						tochnog_element (FP, elem, enodes);

					elemnum[elem]=elem;
					elemgrp[elem]=quadgrp;
					elem++;

				}		/* End of one-slice handling */


				if ((backend==3) && (z==lastz))			/* Create a top mesh of facets */
				{
					stl_plate (FP, sn, xysize*1e-3, nnodes, 0);
				}


			}			/* End z>0 */

			/* Scroll down the nodes from sn to psn */

			if (prevz < 0)
				dp (2, "Storing first-time nodes for use in slice %d\n", z+zskip);
			else
				dp (2, "Storing nodes from slice %d for use in slice %d\n", z, z+zskip);
			for (i=0; i<snodecnt+2; i++)
				psn[i] = sn[i];
			prevz = z;

		}					/* End of if (meshflg[z] )    */
		else
			dp (2, "Skipped empty slice %d\n",z);

	}				/* End of Z loop */

	/* Done writing elements. Now dump the nodes. Note that we need to multiply with 
		pixel size to get coordinates in meters */

	if (backend != 3)			/* Skip this whole section for STL */
	{
		fprintf (FP, "\n");
		for (k=0; k<ncnt; k++)
		{
			fprintf (FP, "node %d    %e  %e  %e\n",k, 
						nodelist[k].x * xysize * 1e-3, 
						nodelist[k].y * xysize * 1e-3, 
						nodelist[k].z * xysize * 1e-3);
		}

		if (press>0)		/* Apply nodal pressure. Note that the parameter is given in MPa, we need to convert to Pa */
		{
			fprintf (FP, "\n\n");

			i=0;
			for (k=0; k<ncnt; k++)
			{
				if (nodelist[k].nodeflags & 0x01)	/* For all interior nodes */
				{
					fprintf (FP, "bounda_unknown %d  %d -pres\n",i,k);
					fprintf (FP, "bounda_time    %d  0. %e 1 %e\n",i, -press*1e6, -press*1e6);
					i++;	
				}
			}
			fprintf (FP, "\n\n");

			for (k=0; k<ncnt; k++)
			{
				if (nodelist[k].nodeflags & 0x02)	/* For all end nodes */
				{
					fprintf (FP, "bounda_unknown %d  %d  -velx -vely -velz\n",i,k);
					fprintf (FP, "bounda_time    %d  0. 0. 1.e6 0.\n",i);
					i++;
				}
			}

			fprintf (FP, "\n\n");

			for (k=0; k<ncnt; k++)
			{
				if (nodelist[k].nodeflags & 0x02)	/* For all end nodes */
				{
					fprintf (FP, "node_boundary %d -yes\n",k);
					i++;	
				}
			}

		}

		/* output the element-group assigment */
		fprintf (FP, "\n\n");
		for (i=0; i<elem; i++)
			fprintf (FP, "element_group    %d  %d\n", elemnum[i], elemgrp[i]);

		/* and, depending on the group population, the material definition lines */

		fprintf (FP, "\n\n");
		for (i=0; i<=groupcnt; i++)
		{
			dp (2, "Group %d has %d elements\n",i,groupocc[i]);
			if (groupocc[i])
				tochnog_material (FP,i, min_young, max_young);		/* Material definition only for groups with nonzero occupancy */
		}
	}
					/* End skipping Tochnog-only stuff */


	if (backend==3)
		stl_footer (FP, 30, backend);
	else
		tochnog_footer (FP, 30, backend);

	fclose (FP);
	free(elemnum); free (elemgrp);
	free (groupocc); free (groupint);
	free (nodelist);
	free (sn); free (psn);
	dp (1, "Meshextract run finished. %d elements and %d nodes found\n",elem,ncnt);

}



/* Integrated mesh extract: Pass1, Pass2, clean-up */


void mod_meshx (image_type *img, float thresh, int numrays, int numinterp, int zskip, int groupcnt, float xysize, float press, float min_young, float max_young, float aniso, const char *filename, char backend, char mode)
{

	meshfx = NULL; meshfy = NULL; meshflg = NULL;

	if (mode==2)				/* Interpolate & Label: Create a secondary image */
	{
		if (allocate_image (&meshx_labelimg, B_RGB, img->xmax, img->ymax, img->zmax)==-1)
			fatal_memerr();
	}

	meshx_pass1 (img, thresh, mode);
	meshx_pass2 (img, thresh, numrays, numinterp, zskip, groupcnt, xysize, press, min_young, max_young, aniso, filename, backend, mode);

	/* Discard arrays */

	free (meshfx); free (meshfy); free (meshflg);

	if (mode==2)
		handback (img, &meshx_labelimg);
}



/**************************************************************************************************************************/

/*          ###    #   #    ###           */
/*         #   #   #   #     #            */
/*         #       #   #     #            */
/*         #  ##   #   #     #            */
/*         #   #   #   #     #            */
/*         #   #   #   #     #            */
/*          ###     ###     ###           */
/*                                        */

int meshx_numrays, meshx_inodes, meshx_zskip, meshx_groups;
float meshx_threshold;
float meshx_press, meshx_aniso, meshx_pixelsize, meshx_max_young, meshx_min_young;
char meshx_filename[1024];
char meshx_backend_function;
char meshx_mode;
GtkWidget *meshx_filename_entry;


void on_meshx_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_meshx_ok_clicked  (GtkButton *button, gpointer user_data)
{

	strncpy (meshx_filename, gtk_entry_get_text (GTK_ENTRY(meshx_filename_entry)), 1020);
	mod_meshx (&mainimg, meshx_threshold, meshx_numrays, meshx_inodes, meshx_zskip, meshx_groups, meshx_pixelsize, meshx_press, meshx_min_young, meshx_max_young, meshx_aniso, meshx_filename, meshx_backend_function, meshx_mode);

	gtk_widget_destroy (GTK_WIDGET(user_data));

}


void meshx_adjustment_changed (GtkWidget* widget, gpointer user_data)
{ 
int which_adj;
float buf;
GtkAdjustment *w;

	
	if (user_data) which_adj = *(int*)user_data; else which_adj=0;
	w = GTK_ADJUSTMENT(widget);			/* Funny way to typecast */

	buf= w->value;
	
	if (which_adj==1)
	{
		meshx_threshold = buf;
		dp (2, "MeshX: Threshold selected is %f\n",meshx_threshold);
	}
	else if (which_adj==2)
	{
		meshx_numrays = buf;
		dp (2, "MeshX: Number of rays now %d\n",meshx_numrays);
	}
	else if (which_adj==3)
	{
		meshx_zskip = buf;
		dp (2, "MeshX: Z slices to skip %d\n",meshx_zskip);
	}
	else if (which_adj==4)
	{
		meshx_press = buf;
		dp (2, "MeshX: Nodal pressure now %f\n",meshx_press);
	}
	else if (which_adj==5)
	{
		meshx_groups = buf;
		dp (2, "MeshX: Number of material groups now %d\n",meshx_groups);
	}
	else if (which_adj==6)
	{
		meshx_aniso = buf;
		dp (2, "MeshX: Z anisotropy now %f\n",meshx_aniso);
	}
	else if (which_adj==7)
	{
		meshx_pixelsize = buf;
		dp (2, "MeshX: XY pixel size now %f mm\n",meshx_pixelsize);
	}
	else if (which_adj==8)
	{
		meshx_min_young = buf;
		dp (2, "MeshX: Lowest Young's modulus set to %f MPa\n",meshx_min_young);
	}
	else if (which_adj==9)
	{
		meshx_max_young = buf;
		dp (2, "MeshX: Highest Young's modulus set to %f MPa\n",meshx_max_young);
	}
	else if (which_adj==10)
	{
		meshx_inodes = buf;
		dp (2, "MeshX: Interpolating %d non-ray nodes\n",meshx_inodes);
	}


}


void meshx_backend_menu (GtkWidget *widget, gpointer user_data)
{
	meshx_backend_function = *(int*) user_data;
}


void meshx_intp_menu (GtkWidget *widget, gpointer user_data)
{
	meshx_mode = *(int*) user_data;
	if (!meshx_mode)
		dp (2, "Nearest-neighbor mode selected\n");
	else if (meshx_mode==1)
		dp (2, "Interpolating mode selected\n");
	else if (meshx_mode==2)
		dp (2, "Interpolate & label mode selected\n");
}



void create_meshx_wnd (image_type* img)
{
  GtkWidget *meshx_wnd;
  GtkWidget *meshx_table1;
  GtkWidget *meshx_label1, *meshx_label2, *meshx_label3, *meshx_label4;
  GtkWidget *meshx_okbutton;
  GtkWidget *meshx_cancelbutton;
  GtkWidget *meshx_t;
  GtkAdjustment *meshx_t_adj;
  GtkWidget *hsep;
  GtkWidget *l2_menu, *l2_menu_menu, *intp_menu, *intp_menu_menu;
  GtkWidget *menuitem;
  GtkWidget *meshx_nrays;
  GtkAdjustment *meshx_nrays_adj;
  GtkWidget *meshx_ninodes;
  GtkAdjustment *meshx_ninodes_adj;
  GtkWidget *meshx_zslice;
  GtkAdjustment *meshx_zslice_adj;
  GtkWidget *meshx_press_slider;
  GtkAdjustment *meshx_press_adj;
  GtkWidget *meshx_group;
  GtkAdjustment *meshx_group_adj;
  GtkWidget *meshx_anisotropy;
  GtkAdjustment *meshx_anisotropy_adj;
  GtkWidget *meshx_pixsize;
  GtkAdjustment *meshx_pixsize_adj;
  GtkWidget *meshx_young1;
  GtkAdjustment *meshx_young1_adj;
  GtkWidget *meshx_young2;
  GtkAdjustment *meshx_young2_adj;
	GtkTooltips *tt_backend, *tt_thresh, *tt_nrays, *tt_zslice, *tt_pixsize, *tt_aniso, *tt_press;
	GtkTooltips *tt_group, *tt_young1, *tt_young2, *tt_fn, *tt_ok, *tt_cancel, *tt_intp, *tt_ninodes;

  
  float min,max;
  int y;

	if (img->imgtype == B_FLOAT) max = img->maxval.fmax; else max = img->maxval.imax;
	if (img->imgtype == B_FLOAT) min = img->minval.fmin; else min = img->minval.imin;
	meshx_threshold = min+1;
	meshx_numrays=64;
	meshx_inodes=0;
	meshx_zskip=1;
	meshx_press=1.0;
	meshx_aniso = 1.0;
	meshx_groups=10;
	meshx_backend_function = 0;
	meshx_pixelsize = 1.0;
	meshx_max_young = 100;
	meshx_min_young = 50;
	meshx_mode = 0;

	meshx_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (meshx_wnd), "meshxes_wnd", meshx_wnd);
	gtk_window_set_title (GTK_WINDOW (meshx_wnd), "Mesh Extraction");
	gtk_container_set_border_width (GTK_CONTAINER (meshx_wnd), 3);
	gtk_window_set_modal (GTK_WINDOW (meshx_wnd), TRUE);

	meshx_table1 = gtk_table_new (17, 2, FALSE);
	gtk_widget_ref (meshx_table1);
	gtk_object_set_data_full (GTK_OBJECT (meshx_wnd), "meshx_table1", meshx_table1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (meshx_table1);
	gtk_container_add (GTK_CONTAINER (meshx_wnd), meshx_table1);

	y=0;
	meshx_label2 = MakeLabel (meshx_wnd, meshx_table1, "Threshold", "meshx_label2", 0,y);
	meshx_t_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_threshold, min, max, 1, 10, 0);
	meshx_t = MakeSpinbutton (GTK_ADJUSTMENT(meshx_t_adj), meshx_wnd, meshx_table1, "meshx_t", 1,y);
	gtk_signal_connect (GTK_OBJECT (meshx_t_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt1);
	tt_thresh = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_thresh, meshx_t, "Segmentation threshold. On each ray, an inner vertex is found\nwhen this value is first exceeded in the image, and an outer\nvertex is found where the image value drops\nbelow this threshold.", NULL);

	y++;
	meshx_label4 = MakeLabel (meshx_wnd,meshx_table1,"Edge Interpolation", "label4",0,y);
	intp_menu = Make_Optionmenu (meshx_wnd,meshx_table1,"meshx_menui",1,y);
	intp_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (intp_menu_menu, meshx_intp_menu,  "Pixel center", (gpointer) &menuopt0);
	menuitem = Option_menuitem (intp_menu_menu, meshx_intp_menu, "Linear interpolation", (gpointer) &menuopt1);
	menuitem = Option_menuitem (intp_menu_menu, meshx_intp_menu, "Interpolate & Label", (gpointer) &menuopt2);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (intp_menu), intp_menu_menu);
	tt_intp = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_intp, intp_menu, "Choose whether edges lie on straight pixels or are interpolated bewteen pixels", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Number of rays", "meshx_label3", 0,y);
	meshx_nrays_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_numrays, 16, 256, 1, 1, 0);
	meshx_nrays = MakeSpinbutton (GTK_ADJUSTMENT(meshx_nrays_adj), meshx_wnd, meshx_table1, "meshx_nrays", 1,y);
	gtk_signal_connect (GTK_OBJECT (meshx_nrays_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt2);
	tt_nrays = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_nrays, meshx_nrays, "Number of probing rays. This value determines the number of\nnon-interpolated nodes per slice.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Number of interpolated nodes", "meshx_label3", 0,y);
	meshx_ninodes_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_inodes, 0, 8, 1, 1, 0);
	meshx_ninodes = MakeSpinbutton (GTK_ADJUSTMENT(meshx_ninodes_adj), meshx_wnd, meshx_table1, "meshx_ninodes", 1,y);
	gtk_signal_connect (GTK_OBJECT (meshx_ninodes_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt10);
	tt_ninodes = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_ninodes, meshx_ninodes, "Number of interpolated nodes created by spline interpolation\nbetween any two non-interpolated nodes.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Z slice factor", "meshx_label4", 0,y);
	meshx_zslice_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_zskip, 1, 16, 1, 1, 0);
	meshx_zslice = MakeSpinbutton (GTK_ADJUSTMENT(meshx_zslice_adj), meshx_wnd, meshx_table1, "meshx_zslice", 1,y);
	gtk_signal_connect (GTK_OBJECT (meshx_zslice_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt3);
	tt_zslice = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_zslice, meshx_zslice, "Number of slices to be skipped. The default value of 1\nmeans that probing rays are emitted in each slice, a value of 2\nmeans that every other slice is skipped etc.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "In-plane pixel size [mm]", "meshx_label7a", 0,y);
	meshx_pixsize_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_pixelsize, 0.01, 10000.0, 0.1, 1.0, 0);
	meshx_pixsize = MakeSpinbutton (GTK_ADJUSTMENT(meshx_pixsize_adj), meshx_wnd, meshx_table1, "meshx_pixsz", 1,y);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON(meshx_pixsize), 3);
	gtk_signal_connect (GTK_OBJECT (meshx_pixsize_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt7);
	tt_pixsize = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_pixsize,  meshx_pixsize, "To convert pixels into real-world units.\nThis value determines the in-plane (x-y) pixel size in mm.\nIn each slice, pixels are assumed isotropic.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Z slice anisotropy", "meshx_label7", 0,y);
	meshx_anisotropy_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_aniso, 1, 64, 1, 1, 0);
	meshx_anisotropy = MakeSpinbutton (GTK_ADJUSTMENT(meshx_anisotropy_adj), meshx_wnd, meshx_table1, "meshx_zaniso", 1,y);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON(meshx_anisotropy), 1);
	gtk_signal_connect (GTK_OBJECT (meshx_anisotropy_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt6);
	tt_aniso = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_aniso, meshx_anisotropy, "Axial anisotropy factor. This value specifies the ratio\nof the voxel size in z to the voxel size in x-y.", NULL);

	y++;
	hsep = MakeHsep (meshx_wnd, meshx_table1, "hseparator4", 0,y);
	hsep = MakeHsep (meshx_wnd, meshx_table1, "hseparator5", 1,y);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Node pressure [MPa]", "meshx_label5", 0,y);
	meshx_press_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_press, 0, 10, 1, 1, 0);
	meshx_press_slider = MakeSpinbutton (GTK_ADJUSTMENT(meshx_press_adj), meshx_wnd, meshx_table1, "meshx_press", 1,y);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON(meshx_press_slider), 3);
	gtk_signal_connect (GTK_OBJECT (meshx_press_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt4);
	tt_press = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_press, meshx_press_slider, "Nodal pressure applied to inner-wall nodes (boundary condition)\n.This is a time-independent pressure and is set in MPa.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Material groups", "meshx_label6", 0,y);
	meshx_group_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_groups, 1, 100, 1, 1, 0);
	meshx_group = MakeSpinbutton (GTK_ADJUSTMENT(meshx_group_adj), meshx_wnd, meshx_table1, "meshx_grp", 1,y);
	gtk_signal_connect (GTK_OBJECT (meshx_group_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt5);
	tt_group = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_group, meshx_group, "Number of material groups. This is the number of\nlinear bins between the lowest and highest image\nvalue per ray, and\neach element is assigned to one of these groups.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Minimum Young's modulus [MPa]", "meshx_label6a", 0,y);
	meshx_young1_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_min_young, 1, 500, 1, 1, 0);
	meshx_young1 = MakeSpinbutton (GTK_ADJUSTMENT(meshx_young1_adj), meshx_wnd, meshx_table1, "meshx_yng1", 1,y);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON(meshx_young1), 2);
	gtk_signal_connect (GTK_OBJECT (meshx_young1_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt8);
	tt_young1 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_young1, meshx_young1, "Youngs modulus (in MPa) for the lowest-valued material group.", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Maximum Young's modulus [MPa]", "meshx_label6b", 0,y);
	meshx_young2_adj = (GtkAdjustment*)gtk_adjustment_new (meshx_max_young, 1, 500, 1, 1, 0);
	meshx_young2 = MakeSpinbutton (GTK_ADJUSTMENT(meshx_young2_adj), meshx_wnd, meshx_table1, "meshx_yng2", 1,y);
	gtk_signal_connect (GTK_OBJECT (meshx_young2_adj), "value_changed", GTK_SIGNAL_FUNC (meshx_adjustment_changed), &menuopt9);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON(meshx_young2), 2);
	tt_young2 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_young2, meshx_young2, "Youngs modulus (in MPa) for the highest-valued material group.", NULL);

	y++;
	hsep = MakeHsep (meshx_wnd, meshx_table1, "hseparator4", 0,y);
	hsep = MakeHsep (meshx_wnd, meshx_table1, "hseparator5", 1,y);

	y++;
	meshx_label1 = MakeLabel (meshx_wnd,meshx_table1,"Tochnog Backend", "label1",0,y);
	l2_menu = Make_Optionmenu (meshx_wnd,meshx_table1,"meshx_menu1",1,y);
	l2_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (l2_menu_menu, meshx_backend_menu,  "GiD (default)", (gpointer) &menuopt0);
	menuitem = Option_menuitem (l2_menu_menu, meshx_backend_menu, "VTK (Paraview)", (gpointer) &menuopt1);
	menuitem = Option_menuitem (l2_menu_menu, meshx_backend_menu, "OpenDX", (gpointer) &menuopt2);
	menuitem = Option_menuitem (l2_menu_menu, meshx_backend_menu, "STL/ASCII", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (l2_menu), l2_menu_menu);
	tt_backend = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_backend, l2_menu, "This directs Tochnog to provide output in one of three\navailable formats, GiD, VTK (for Paraview) and OpenDX", NULL);

	y++;
	meshx_label3 = MakeLabel (meshx_wnd, meshx_table1, "Tochnog file", "meshx_label6", 0,y);
	meshx_filename_entry = Entry_in_table (meshx_wnd, meshx_table1, "meshextract.txt", "meshx_entry", 1,y);
	tt_fn = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_fn, meshx_filename_entry, "Output file name (this is the Tochnog input file).\nThis file is a plain ASCII file that contains all\nelements, nodes, material properties, boundary conditions\nand Tochnog control statements.", NULL);

	y++;
	hsep = MakeHsep (meshx_wnd, meshx_table1, "hseparator4", 0,y);
	hsep = MakeHsep (meshx_wnd, meshx_table1, "hseparator5", 1,y);

	y++;
	meshx_okbutton = Button_in_table (meshx_wnd, meshx_table1, "OK", "meshx_okbutton", 0,y, NULL);
	meshx_cancelbutton = Button_in_table (meshx_wnd, meshx_table1, "Cancel", "meshx_canbutton", 1,y, NULL);
	tt_ok = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_ok, meshx_okbutton, "Press OK to start mesh extraction.", NULL);
	tt_cancel = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_cancel, meshx_cancelbutton, "Press Cancel to abort and close this window.", NULL);



	gtk_signal_connect (GTK_OBJECT (meshx_okbutton), "clicked", 
                    	GTK_SIGNAL_FUNC (on_meshx_ok_clicked), meshx_wnd);
	gtk_signal_connect (GTK_OBJECT (meshx_cancelbutton), "clicked", 
                    	GTK_SIGNAL_FUNC (on_meshx_cancel_clicked), meshx_wnd);
	gtk_signal_connect (GTK_OBJECT (meshx_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_meshx_cancel_clicked), meshx_wnd);


	gtk_widget_show (meshx_wnd);
}




/**************************************************************************************************************************/

/*         ####     ###     ###             ###    #####   ####    #   #    ###    #####   #   #   ####    #####          */
/*         #   #   #   #   #   #           #   #     #     #   #   #   #   #   #     #     #   #   #   #   #              */
/*         #   #   #       #   #           #         #     #   #   #   #   #         #     #   #   #   #   #              */
/*         #   #    ###    #   #            ###      #     ####    #   #   #         #     #   #   ####    ####           */
/*         #   #       #   #   #               #     #     # #     #   #   #         #     #   #   # #     #              */
/*         #   #   #   #   #   #           #   #     #     #  #    #   #   #   #     #     #   #   #  #    #              */
/*         ####     ###     ###             ###      #     #   #    ###     ###      #      ###    #   #   #####          */
/*                                                                                                                        */




/**************************************************************

	This is the infrastructure we use to register with the 
	main program. The pointer "menufunction" points to the 
	main entry (the function to be called when the menu is 
	activated). Initmodule is the one function that the 
	main program calls (the name is fixed!) to initialize
	the module and to register the module with the menu.

***************************************************************/

void (*menufunction1)();		/* A pointer to the menu callback function for the main menu */




/* This function corresponds to the driver functions in callbacks.c. Its only
	purpose is to call the actual operator.
	For multiple operators, each operator has one menu function pointer and one of
	these driver functions */

void module_meshx_driver ()
{

	if (file_ok) 
	{
		create_meshx_wnd (&mainimg);		/* Execute the module's operator with mainimage */
	}
}




void initmodule ()
{

	dp (1, "Meshextract: Initializer called\n");
	
	/* Register this module with the main menu */
	
	menufunction1 = &module_meshx_driver;
	dso_menu_register ("Mesh Extract", menufunction1);


	
}

