

/*
 *	macro.c: Macro interpreter and ancilliary functions
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>


#include "application.h"
#include "dlgbox.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "import.h"


#include "macro2.h"



/*****************************************************************

	Here are the functions for flow of control (FOR, WHILE...)
	
	All functions obtain the buffer
	and an index to the first character of the parameter list
	(i.e. just after the opening parentheses). They are responsible
	for getting the required parameters and advancing the index
	to past the closing parentheses (so it now points to the ;)
	
	Note that any procedure or function that takes an image expr
	as parameter is responsible for disposing of the image buffer-
	expr allocates a "dangling" image that needs to be passed thru
	(as return variable) or needs to be freed. In many cases, the
	actual function performs this job. Sometimes (particularly in
	cases where the return type is MT_REAL) it needs to be
	freed inside the function.
	
******************************************************************/


/* Search for a matching token in case of a rejected loop.
   For example, a while loop that begins with a FALSE condition
   will never be entered, so we need to find the matching
   ENDWHILE.
   toktype determines the token pairs as follows:
   0: FOR/FOREACH and ENDFOR
   1: WHILE and ENDWHILE
   2: IF and ENDIF (ignoring ELSE)
   3: IF and ELSE/ENDIF (for the ELSE part of the IF statement)
   
   Note that we don't modify the index, but rather return the position

   BUG: This function gets confused with nested IF-ELSE-ENDIF 

*/


int forward_search_token (unsigned char* s, int* idx, int toktype)
{
int i,level;
int uptoken1, uptoken2, downtoken1,downtoken2;
char strdummy[1024];

	i = *idx;

	
	uptoken1=0; uptoken2=0; downtoken1=0; downtoken2=0;
	if (toktype==0)
	{
		uptoken1=132; uptoken2=133; downtoken1=131;
	}
	else if (toktype==1)
	{
		uptoken1=138; downtoken1=137;
	}
	else if (toktype==2)
	{
		uptoken1=136; downtoken1=134;
	}
	else if (toktype==3)
	{
		uptoken1=136; downtoken1=134; downtoken2=135;
	}
	
	/* assume that i indexes the first statement after the uptoken */
	
	level=1;
	while ((level>0) && (i < strlen(s)))
	{
		i++;
		if (s[i]=='"')	/* we must skip whatever is in quotes */
			extract_string (s, &i, strdummy, 1023);
		else if (s[i]==downtoken1)
			level--;
		else if ((downtoken2!=0) && (s[i]==downtoken2))
			level--;
		else if (s[i]==uptoken1)
			level++;
		else if ((uptoken2!=0) && (s[i]==uptoken2))
			level++;
		else if ((s[i]==128) || (s[i]==129))	/* Can't run into ENDMACRO or ENDPROC */
			macro_error (24,i);		
	}
	if (level>0) macro_error (24,i);
	
	/* position *on* the ; */
	if (s[++i] != ';') macro_error (5,i);
	
	return i;
}




/*
	endmacro procedure
	syntax:	endmacro;
	No parameters.
	Endmacro disposes of all variables, displays the log buffer,
	and returns control to the menu
*/



variable_type mac_endmacro (unsigned char* s, int* idx)
{
int i;

	/* If we end top level macro, we need to free the variable list */
	/* KLUDGE! In some instances, endmacro is called twice, which is not good
		since everything is already freed. So we test if macrorun is already 0
		and yell a warning if yes. */
	
	if (!macrolevel)
	{
		if (macrorun==0)
			dp (0, "WARNING: Executing ENDMACRO after macro run has finished.\n");
		free_all_vars ();
		macrorun=0;
		if (mac_has_output) 
		{
			ShowLongMessage (mac_logbuffer, "Macro results");
			free (mac_logbuffer); 
			mac_logbuffer = NULL;
		}
		mac_has_output = 0;		/* Same kludge. Avoid a double-free */
		dp (1, "Macro");
	}
	else
		dp (1, "Submacro");

	if ((s==NULL) && (idx==0))
		dp (1, " finished with error(s)\n");
	else
		dp (1, " finished successfully\n");
		
	return nullvar;
}



/*
	call procedure
	syntax:	call ("Procedure");
	Parameter: Name of the procedure or macro to execute
	This executes a macro or procedure and returns control
	to the calling program.

*/


variable_type mac_callproc (unsigned char* master, int* idx)
{
int macronum,i,j,k, rtncode;
char submacro[128];
unsigned char* buf;
variable_type r;


	/* Grab the macro name from the calling stream */
	
	j = *idx;
	r = expr (master, &j, MT_STRING);
	strncpy (submacro, r.strg, 127);
	free (r.strg);
	/* extract_string (master, &j, submacro, 127); */

	/* Find the macro/procedure in macro table */
	
	i=0; macronum=-1;
	while ((i<nummacros) && (macronum==-1))
	{
		if ((strlen(submacro)==strlen(macronames[i]))
			&& (!strncasecmp (submacro, macronames[i], 128)))
			macronum=i;
		else
			i++;
	}
	if (macronum==-1) macro_error (18,j);
	
	/* Skip to closing parentheses, update pointer */

	while (master[j]!=')') j++;
	*idx = j+1;

	macrolevel++;
	buf = macrodef[macronum];
	dp (1, "Executing submacro %d: %s\n",macronum,(char*)macronames[macronum]);
	i=0;
	do
	{
		j=i+1; k=0; current_macro_pos = i;
		while ((j<strlen(buf)) && (buf[j]!=';')) j++;
		listbuf (buf,i,j,current_macro,2040);
		
		rtncode=exec_macro_single (buf, &i);
	}
	while ( (i<strlen(buf)) && (rtncode==0));

	macrolevel--;

	dp (1, "\n");
	return nullvar;

}




/*
	for directive
	Syntax:	for (numvar = expr1, expr2); instructions; endfor;
			for (numvar = expr1, expr2, expr3); instructions; endfor;
	
	This is an implementation of the for loop. The for-control
	variable, numvar, is set to expr1 at the beginning. The end value
	is expr2, and the optional increment is given by expr3.
	Note that the last value, numvar==expr2, will be executed.
*/


variable_type mac_forloop (unsigned char* master, int* idx)
{
variable_type v,r;
int j,k, endfor_token, loopstart, loopend;
float start, end, increment;
long istart, iend, iincrement;
char intloop;

	j=*idx; k=0;

	/* Since the tokens may vary, we need to grab the FOR token.
	   This allows us to identify the ENDFOR token farther down.
	   Note the order in the keywordlist: ENDFOR - FOREACH - FOR */

	endfor_token = master[j-2] - 2;
	if (!isalpha(master[j])) macro_error (1,j);	
	v = nullvar;
	
	do
	{
		v.name[k++] = master[j++];
	} 
	while ((k<16) && (isalpha2(master[j])));
	v.name[k]=0;

	/* Check for the = token, else synerr */
	
	if (master[j++] != 130) macro_error (1,j);
	
	/* Now get the two or three expressions. Note that there are two
	   different versions, the integer loop and the real loop. */
	
	increment = 1.0; iincrement=1; intloop=0; end=0.0; iend=0;
	r = expr (master,&j,MT_NULL); 
	
	/* This decides whether int or real. Get all three params,
	   start, end and increment; and also set the initial condition */
	
	if (r.type==MT_REAL)
	{
		start = r.numvalue; intloop=0;
		if (master[j++]!=',') macro_error (1,j);
		r = expr (master,&j,MT_REAL); end = r.numvalue;
		if (master[j]==',')
		{
			j++; 
			r = expr (master,&j,MT_REAL); increment = r.numvalue;
		}
		v.numvalue = start; v.type = MT_REAL; storevar (v);
		dp (2, "FOR-loop level %d, control %s starts at %f up to %f by %f\n",
					forlevel, v.name, v.numvalue, end, increment);
	
	}
	else if (r.type==MT_INT)
	{
		istart = r.intvalue; intloop=1;
		if (master[j++]!=',') macro_error (1,j);
		r = expr (master,&j,MT_INT); iend = r.intvalue;
		if (master[j]==',')
		{
			j++; 
			r = expr (master,&j,MT_INT); iincrement = r.intvalue;
		}
		v.intvalue = istart; v.type = MT_INT; storevar (v);
		dp (2, "FOR-loop level %d, control %s starts at %li up to %li by %li\n",
					forlevel, v.name, v.intvalue, iend, iincrement);
	}
	else macro_error (6,j);
	
	while (master[j]!=')') j++;
	
	/* Since we are now in control, we need to skip the ';' to
	   reach the next instruction. In addition, save the pointer
	   value as loop start for the iterations */
	   
	if (master[++j] != ';') macro_error (1,j);
	forlevel++; j++; loopstart = j;
	loopend = loopstart+1; /* Silence a compiler warning */
	
	
	/* Here is the actual loop; a call to exec_macro_single. */
	
	if (intloop)
	{
		while (((iincrement >=0) && (v.intvalue<=iend)) || ((iincrement <0) && (v.intvalue>=iend)))
		{	
			while (master[j]!=endfor_token)
			{
				exec_macro_single (master, &j);
			}
			loopend = j;
			dp (2, "Loop level %d. Run into ENDFOR at %d\n",forlevel,j);

			/* Update the control variable and reset the buffer pointer */

			v.intvalue += iincrement; storevar (v);
			j = loopstart;
		}
	}
	else
	{
		while (((increment >=0) && (v.numvalue<=end)) || ((increment <0) && (v.numvalue>=end)))
		{	
			while (master[j]!=endfor_token)
			{
				exec_macro_single (master, &j);
			}
			loopend = j;
			dp (2, "Loop level %d. Run into ENDFOR at %d\n",forlevel,j);

			/* Update the control variable and reset the buffer pointer */

			v.numvalue += increment; storevar (v);
			j = loopstart;
		}
	}
	j = loopend;
	
	/* j points at the endfor token, skip it, but leave the subsequent ';'  */
	
	*idx = j+1;
	forlevel--;

	return nullvar;
}

	

/*
	foreach directive
	Syntax:	foreach (strvar, "mask"); instructions; endfor;
			foreach (strvar, "mask", "cwd"); instructions; endfor;
	
	Foreach is a special case of while-loop which repeats a set of
	instructions for each regular file in the current working directory
	that matches "mask". Optinally, the current working directory "cwd"
	may be set.

	TODO: This function fails with a useless error message if there are no files
*/


variable_type mac_foreach (unsigned char* master, int* idx)
{
variable_type v,r;
int j,k, endfor_token, loopstart, loopend, filecnt, err;
char mask[256];
char cwd[1024];
char resfile[2048];
struct stat statresult;		/* to check if a file is a directory */
DIR* directory;
struct dirent* entry;

	j=*idx; k=0;
	memset (mask, 0, 256);
	memset (cwd, 0, 1024);

	/* Since the tokens may vary, we need to grab the FOREACH token.
	   This allows us to identify the ENDFOR token farther down.
	   Note the order in the keyword list: ENDFOR - FOREACH - FOR */

	endfor_token = master[j-2] - 1;
	if (!isalpha(master[j])) macro_error (1,j);	
	v = nullvar;
	
	do
	{
		v.name[k++] = master[j++];
	} 
	while ((k<16) && (isalpha2(master[j])));
	v.name[k]=0;
	v.type=MT_STRING;

	/* Get the mask */
	
	if (master[j++] != ',') macro_error (1,j);
	r = expr (master,&j,MT_STRING);
	strncpy (mask,r.strg,255); free (r.strg);
	
	/* Get the optional cwd */
	
	if (master[j]==',')
	{
		j++;
		r = expr (master,&j,MT_STRING);
		strncpy (cwd,r.strg,1023); free (r.strg);
	}

	/* Skip to closing parentheses and save location as loop start */
	
	while (master[j]!=')') j++;
	if (master[++j] != ';') macro_error (1,j);
	forlevel++; j++; loopstart = j;
	loopend = loopstart+1; /* Silence a compiler warning */
	
	/* We got everything collected. Start preparing the directory scan.
		Note that opendir returns NULL in case of error, which we have to trap. */
	
	if (strlen(cwd))	
		directory = opendir (cwd);
	else
		directory = opendir ("./");
	if (!directory) macro_error (29, j);
	
	filecnt=0;
	do
	{
		entry = readdir (directory);
		if (entry) 
		{
			if (strlen(cwd))
			{
				strcpy (resfile, cwd); strcat (resfile, "/");
			}
			else
				strcpy (resfile, "./");

			strcat (resfile, entry->d_name);
			err=stat (resfile, &statresult);
			if ( (!S_ISDIR (statresult.st_mode))	/* Skip directories */
				&& (!err)							/* error STATing */
				&& (shmatch (mask, entry->d_name))) /* matching the mask */
			{
				filecnt ++;
				
				/* Store the current file name in the control variable */
				
				v.strg = calloc (strlen (resfile)+1, sizeof(char));
				strcpy (v.strg, resfile);
				storevar (v);
				
				dp (2, "FOREACH: File %d, %s\n",filecnt,v.strg);
				
				/* Execute single commands until ENDFOR */
				
				while (master[j]!=endfor_token)
				{
					exec_macro_single (master, &j);
				}
				loopend = j;
				dp (2, "Loop level %d. Run into ENDFOR at %d\n",forlevel,j);

				/* Reset the buffer pointer for next loop iteration */

				j = loopstart;
			}
		}
	}
	while (entry);
	
	if (filecnt==0)		/* Nothing found, skip over the loop */
	{
		loopend = forward_search_token (master, &j, 0);
		dp (2, "Rejected loop. Found ENDFOR at %d\n",loopend-1);
		*idx=loopend;
	}
	else
		*idx = loopend+1;

	forlevel--;

	return nullvar;
}





/*
	while directive
	Syntax:	while (intexpr); instructions; endwhile;
	
	This is an implementation of the while loop.
	The instructions will be executed as long as intexpr
	is nonzero (boolean TRUE). If intexpr is zero when the loop
	is first entered, instructions will not be executed at all.
*/


variable_type mac_while (unsigned char* master, int* idx)
{
variable_type r;
int j,k, cond, loopstart, loopend, endwhile_token, loopcnt;

	j=*idx; k=0; loopcnt=0;
	loopstart = j;		/* Now first char of intexpr */
	loopend=j;
	endwhile_token = master[j-2] - 1;
	
	dp (2, "WHILE loop starts at %d\n",j);
	
	do
	{
		r = expr (master,&j,MT_INT);
		cond = r.intvalue;
		while (master[j]!=')') j++;	/* The boolean expr closes with parentheses... */
		if (master[++j] != ';') macro_error (1,j);	/* ...followed by a semicolon */
		j++;		/* skip over the semicolon */
		
		if (cond)
		{	
			loopcnt++;
			while (master[j]!=endwhile_token)
			{
				exec_macro_single (master, &j);
			}
			loopend = j;
			dp (2, "Run into ENDWHILE at %d\n",j);
			j = loopstart;	/* reset the loop pointer to point at the bool expr */
		}
	}
	while (cond);

	if (loopcnt==0)		/* never true, need to find ENDWHILE */
	{
		loopend = forward_search_token (master, &j, 1);
		dp (2, "Rejected loop. Found ENDWHILE at %d\n",loopend-1);
		*idx=loopend;
	}
	else
		*idx = loopend+1;

	return nullvar;
}




/*
	repeat directive
	Syntax:	repeat; instructions; until (intexpr);
	
	This is an implementation of the repeat/until loop.
	The instructions will be executed at least once and
	as long as intexpr is zero.
*/


variable_type mac_repeat (unsigned char* master, int* idx)
{
variable_type r;
int j,k, cond, loopstart, until_token, loopcnt;

	j=*idx; k=0; loopcnt=0;
	j++;					/* skip over ; */
	loopstart = j;			/* First char of instruction block */
	until_token = master[j-2] - 1;
	
	dp (2, "REPEAT loop starts at %d\n",loopstart);
	
	do
	{
		loopcnt++;
		while (master[j]!=until_token)
		{
			exec_macro_single (master, &j);
		}
		
		/* j points at the until-token. Skip over the '(' and evaluate expr. */
		
		if (master[++j]!='(') macro_error (4,j);
		j++;
		r = expr (master,&j,MT_INT);
		cond = r.intvalue;
		while (master[j]!=')') j++;	/* The boolean expr closes with parentheses... */
		if (master[++j] != ';') macro_error (4,j);	/* ...followed by a semicolon */
		j++;		/* skip over the semicolon */

		dp (2, "REPEAT loop ends at %d\n",j);
		if (!cond) j = loopstart;	/* reset the loop pointer to point at the bool expr */

	}
	while (!cond);

	*idx = j-1;

	return nullvar;
}



/*
	if directive
	Syntax:	if (intexpr); 
				instructions1;
			[else;
				instructions2;]
			endif;
	
	This is an implementation of the if/then/else branching instruction.
	If intexpr is nonzero, the block instructions1 will be executed,
	otherwise the block instructions2. Else and instructions2 are optional.
	
*/


variable_type mac_if (unsigned char* master, int* idx)
{
variable_type r;
int cond,j,k,else_token,endif_token;

	j=*idx; 
	else_token = master[j-2] - 1;
	endif_token = master[j-2] - 2;
	iflevel++;
	
	dp (2, "IF block, level %d, starts at %d\n",iflevel,j);
	
	r = expr (master,&j,MT_INT);
	cond = r.intvalue;
	while (master[j]!=')') j++;	/* The boolean expr closes with parentheses... */
	if (master[++j] != ';') macro_error (1,j);	/* ...followed by a semicolon */
	j++;		/* skip over the semicolon */

	if (cond)	/* Execute until "else" found, skip over "else" block */
	{
		dp (2, "IF: Level %d: Entering TRUE block\n",iflevel);
		while ((master[j]!=else_token) && (master[j]!=endif_token))
		{
			exec_macro_single (master, &j);
		}
		
		/* If there is an "else", skip over it. */
		
		if (master[j]==else_token)
		{
			dp (2, "IF: Level %d: Skipping ELSE section\n",iflevel);
			if (master[++j]!=';') macro_error (4,j);
			j++;
			k=forward_search_token (master,&j,2);
			j=k;
		}
		else
			j++;
	}
	else	/* Skip until "else" _or_ "endif" found */
	{
		dp (2, "IF: Level %d: Entering FALSE block\n",iflevel);
		k=forward_search_token (master,&j,3);
		j=k; k--;		/* K to peek at the token, j *on* the ; */

		if (master[k]==else_token)	/* Skip this part if no "else" exists */
		{
			dp (2, "IF: Level %d: Skipping - no ELSE section\n",iflevel);
			j++;
			while (master[j]!=endif_token)
			{
				exec_macro_single (master, &j);
			}
			if (master[++j] != ';') macro_error (4,j);
		}
	
	}

	*idx = j;
	iflevel--;
	return nullvar;
}




/*****************************************************

######
#     #   #   #    ##    #####      #     ####
#     #    # #    #  #   #    #     #    #    #
#     #     #    #    #  #    #     #    #
#     #     #    ######  #    #     #    #
#     #     #    #    #  #    #     #    #    #
######      #    #    #  #####      #     ####

*****************************************************/


/* These operators are fairly complex because of the different
   type combinations. As a general rule, this table tells the
   resulting type for x type (columns) and y type (rows):
   
			int			real		image		string

int:		int			real		image(1)	n.a.
real:		real		real		image(1)	n.a.
image:		n.a.		n.a.		image(2)	n.a.
string:		n.a.		n.a.		n.a.		string

(1): Not available for Boolean operators.
Note: Whenever we mix an image type with any numerical type
(either MT_REAL or MT_INT), the result image is B_FLOAT;
(2) This is the exception. Boolean operators are invalid
for B_FLOAT images, but retain the image type for all others.
Note: Comparison operators are invalid for images; they always
return integer values. Comparisons are valid for strings.
*/

variable_type mac_add (variable_type x, variable_type y)
{
variable_type v;
float* p;
float f;

int X,Y,Z; long idx;

	v = nullvar;
	v.type = x.type; 		/* Default. Reset only if needed */
	if ((x.type==MT_INT) && (y.type==MT_INT))
		v.intvalue = x.intvalue + y.intvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
		v.numvalue = x.numvalue + y.numvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
	{
		v.numvalue = x.numvalue + y.intvalue;
		v.type = MT_REAL;
	}
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
	{
		v.numvalue = x.intvalue + y.numvalue;
		v.type = MT_REAL;
	}
	else if ((x.type == MT_IMG) && (y.type==MT_IMG))
	{
		dp (2, "Addition of two images\n");
		if (!img_match_size (&x.imgptr, &y.imgptr)) macro_error (12,current_macro_pos);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) + readbuf_flt(&y.imgptr,X,Y,Z);
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
		freebuf (&y.imgptr);
	
	}
	else if ((x.type == MT_IMG) && ((y.type==MT_REAL) || (y.type==MT_INT)))
	{
		if (y.type==MT_REAL) f=y.numvalue; else f=y.intvalue;
		dp (2, "Addition of constant %f to image\n",f);
		
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) + f;
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
	
	}
	else if ((y.type == MT_IMG) && ((x.type==MT_REAL) || (x.type==MT_INT)))
	{
		if (x.type==MT_REAL) f=x.numvalue; else f=x.intvalue;
		dp (2, "Addition of constant %f to image\n",f);
		
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, y.imgptr.xmax, y.imgptr.ymax,y.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&y.imgptr,X,Y,Z) + f;
		make_minmax (&v.imgptr);
		freebuf (&y.imgptr);
	
	}
	else if ((x.type==MT_STRING) && (y.type==MT_STRING))
	{
		char* p;
		p = calloc (strlen(x.strg)+strlen(y.strg)+1, sizeof(char));
		strcat (p, x.strg);
		strcat (p, y.strg);
		v.type = MT_STRING;
		v.strg = p;
		free (x.strg); free (y.strg);
	}
	else
		macro_error (6,current_macro_pos);
	
	return v;
}



variable_type mac_subtract (variable_type x, variable_type y)
{
variable_type v;
float* p;
float f;

int X,Y,Z; long idx;

	v = nullvar;
	v.type = x.type; 		/* Default. Reset only if needed */
	if ((x.type==MT_INT) && (y.type==MT_INT))
		v.intvalue = x.intvalue - y.intvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
		v.numvalue = x.numvalue - y.numvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
	{
		v.numvalue = x.numvalue - y.intvalue;
		v.type = MT_REAL;
	}
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
	{
		v.numvalue = x.intvalue - y.numvalue;
		v.type = MT_REAL;
	}
	else if ((x.type == MT_IMG) && (y.type==MT_IMG))
	{
		dp (2, "Subtraction of two images\n");
		if (!img_match_size (&x.imgptr, &y.imgptr)) macro_error (12,current_macro_pos);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z= 0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) - readbuf_flt(&y.imgptr,X,Y,Z);
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
		freebuf (&y.imgptr);
	
	}
	else if ((x.type == MT_IMG) && ((y.type==MT_REAL) || (y.type==MT_INT)))
	{
		if (y.type==MT_REAL) f=y.numvalue; else f=y.intvalue;
		dp (2, "Subtract constant %f from image\n",f);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) - f;
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
	
	}
	else if ((y.type == MT_IMG) && ((x.type==MT_REAL) || (x.type==MT_INT)))
	{
		if (x.type==MT_REAL) f=x.numvalue; else f=x.intvalue;
		dp (2, "Subtract image from constant %f\n",f);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, y.imgptr.xmax, y.imgptr.ymax,y.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = f - readbuf_flt (&y.imgptr,X,Y,Z);
		make_minmax (&v.imgptr);
		freebuf (&y.imgptr);
	
	}
	else
		macro_error (6,current_macro_pos);
	
	return v;
}


variable_type mac_mult (variable_type x, variable_type y)
{
variable_type v;
float* p;
float f;

int X,Y,Z; long idx;

	v = nullvar;
	v.type = x.type; 		/* Default. Reset only if needed */
	if ((x.type==MT_INT) && (y.type==MT_INT))
		v.intvalue = x.intvalue * y.intvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
		v.numvalue = x.numvalue * y.numvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
	{
		v.numvalue = x.numvalue * y.intvalue;
		v.type = MT_REAL;
	}
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
	{
		v.numvalue = x.intvalue * y.numvalue;
		v.type = MT_REAL;
	}
	else if ((x.type == MT_IMG) && (y.type==MT_IMG))
	{
		dp (2, "Multiplication of two images\n");
		if (!img_match_size (&x.imgptr, &y.imgptr)) macro_error (12,current_macro_pos);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z= 0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) * readbuf_flt(&y.imgptr,X,Y,Z);
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
		freebuf (&y.imgptr);
	
	}
	else if ((x.type == MT_IMG) && ((y.type==MT_REAL) || (y.type==MT_INT)))
	{
		if (y.type==MT_REAL) f=y.numvalue; else f=y.intvalue;
		dp (2, "Multiply image with constant %f\n",f);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) * f;
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
	
	}
	else if ((y.type == MT_IMG) && ((x.type==MT_REAL) || (x.type==MT_INT)))
	{
		if (x.type==MT_REAL) f=x.numvalue; else f=x.intvalue;
		dp (2, "Multiply image with constant %f\n",f);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, y.imgptr.xmax, y.imgptr.ymax,y.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&y.imgptr,X,Y,Z) * f;
		make_minmax (&v.imgptr);
		freebuf (&y.imgptr);
	
	}
	else
		macro_error (6,current_macro_pos);
	
	return v;
}


variable_type mac_div (variable_type x, variable_type y)
{
variable_type v;
float* p;
float f;
int X,Y,Z; long idx;

	v = nullvar;
	v.type = x.type; 		/* Default. Reset only if needed */
	if ((x.type==MT_INT) && (y.type==MT_INT))
	{
		if (y.intvalue ==0) macro_error (13, current_macro_pos);
		v.intvalue = x.intvalue / y.intvalue;
	}
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
	{
		if (y.numvalue ==0.0) macro_error (13, current_macro_pos);
		v.numvalue = x.numvalue / y.numvalue;
	}
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
	{
		if (y.intvalue ==0) macro_error (13, current_macro_pos);
		v.numvalue = x.numvalue / y.intvalue;
		v.type = MT_REAL;
	}
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
	{
		if (y.numvalue ==0.0) macro_error (13, current_macro_pos);
		v.numvalue = x.intvalue / y.numvalue;
		v.type = MT_REAL;
	}
	else if ((x.type == MT_IMG) && (y.type==MT_IMG))
	{
		dp (2, "Division of two images\n");
		if (!img_match_size (&x.imgptr, &y.imgptr)) macro_error (12,current_macro_pos);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		/* No division by zero error here, but let any pixel p/0.0 be 0.0.as well */
		
		idx=0;
		for (Z= 0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
				{
					f = readbuf_flt(&y.imgptr,X,Y,Z);
					if (f==0.0) 
						p[idx++]=0.0;
					else
						p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) /f;
				}
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
		freebuf (&y.imgptr);
	
	}
	else if ((x.type == MT_IMG) && ((y.type==MT_REAL) || (y.type==MT_INT)))
	{
		if (y.type==MT_REAL) f=y.numvalue; else f=y.intvalue;
		dp (2, "Divide image by constant %f\n",f);
		if (f ==0.0) macro_error (13, current_macro_pos);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
					p[idx++] = readbuf_flt (&x.imgptr,X,Y,Z) / f;
		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
	
	}
	else if ((y.type == MT_IMG) && ((x.type==MT_REAL) || (x.type==MT_INT)))
	{
		if (x.type==MT_REAL) f=x.numvalue; else f=x.intvalue;
		dp (2, "Divide constant %f by image\n",f);
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, B_FLOAT, y.imgptr.xmax, y.imgptr.ymax,y.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		p = v.imgptr.data;
		
		idx=0;
		for (Z=0; Z<v.imgptr.zmax; Z++)
			for (Y=0; Y<v.imgptr.ymax; Y++)
				for (X=0; X<v.imgptr.xmax; X++)
				{
					float imgval = readbuf_flt (&y.imgptr,X,Y,Z);
					if (abs(imgval)<1e-12) p[idx++]=0.0; else p[idx++] = f/imgval;
				}
		make_minmax (&v.imgptr);
		freebuf (&y.imgptr);
	
	}
	else
		macro_error (6,current_macro_pos);
	
	return v;
}


/* Boolean operators.
   ------------------
   
   These are only defined for integer-based images and integer scalars


*/


#define GENERAL_BOOL(X,Y,FN)	( (FN)==0? (X)&(Y) : (FN)==1? (X)|(Y) : (X)^(Y))


variable_type mac_bool (variable_type x, variable_type y, int function)
{
variable_type v;
long idx,elems;
int op;

	v = nullvar;
	if ((x.type==MT_INT) && (y.type==MT_INT))
	{
		v.intvalue = x.intvalue & y.intvalue;
		v.type = MT_INT;
	}
	else if ((x.type == MT_IMG) && (y.type==MT_IMG))
	{
		dp (2, "Boolean operation on two images\n");
		if (!img_match_size (&x.imgptr, &y.imgptr)) macro_error (12,current_macro_pos);
		if (x.imgptr.imgtype!=y.imgptr.imgtype) macro_error (17, current_macro_pos);
		if (x.imgptr.imgtype==B_FLOAT) macro_error (17, current_macro_pos);
		
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, x.imgptr.imgtype, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		
		elems = x.imgptr.xmax * x.imgptr.ymax * x.imgptr.zmax;
		
		/* The only cases left are short and char/rgb */
		
		if (x.imgptr.imgtype==B_SHORT)
		{
			short* p; short* px; short* py;
			p = v.imgptr.data; px=x.imgptr.data; py=y.imgptr.data;

			for (idx=0; idx < elems; idx++)
				p[idx] = GENERAL_BOOL (px[idx], py[idx], function);
		}
		else if ((x.imgptr.imgtype==B_CHAR) || (x.imgptr.imgtype==B_RGB))
		{
			unsigned char* p; unsigned char* px; unsigned char* py;
			p = v.imgptr.data; px=x.imgptr.data; py=y.imgptr.data;
			if (x.imgptr.imgtype==B_RGB) elems*=3;

			for (idx=0; idx < elems; idx++)
				p[idx] = GENERAL_BOOL (px[idx], py[idx], function);
		}

		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
		freebuf (&y.imgptr);
	
	}
	else if ( ((x.type == MT_IMG) && (y.type==MT_INT)) || ((x.type == MT_INT) && (y.type==MT_IMG)) )
	{
		dp (2, "Boolean operation on image image with scalar\n");
		if (y.type==MT_IMG) macro_swapvar (&x, &y);	/* make sure y is the scalar */
		if (x.imgptr.imgtype==B_FLOAT) macro_error (17, current_macro_pos);
		
		v.type = MT_IMG;
		if (allocate_image (&v.imgptr, x.imgptr.imgtype, x.imgptr.xmax, x.imgptr.ymax,x.imgptr.zmax)==-1)
			macro_error (11,current_macro_pos);
		
		elems = x.imgptr.xmax * x.imgptr.ymax * x.imgptr.zmax;
		op = y.intvalue;
		
		/* The only cases left are short and char/rgb */
		
		if (x.imgptr.imgtype==B_SHORT)
		{
			short* p; short* px;
			p = v.imgptr.data; px=x.imgptr.data;

			for (idx=0; idx < elems; idx++)
				p[idx] = GENERAL_BOOL (px[idx], op, function);
		}
		else if ((x.imgptr.imgtype==B_CHAR) || (x.imgptr.imgtype==B_RGB))
		{
			unsigned char* p; unsigned char* px;
			p = v.imgptr.data; px=x.imgptr.data;
			if (x.imgptr.imgtype==B_RGB) elems*=3;

			for (idx=0; idx < elems; idx++)
				p[idx] = GENERAL_BOOL (px[idx], op, function);
		}

		make_minmax (&v.imgptr);
		freebuf (&x.imgptr);
		
	}
	else
		macro_error (6,current_macro_pos);

	return v;
}


variable_type mac_and (variable_type x, variable_type y)
{
	return mac_bool (x,y,0);
}

variable_type mac_or (variable_type x, variable_type y)
{
	return mac_bool (x,y,1);
}

variable_type mac_xor (variable_type x, variable_type y)
{
	return mac_bool (x,y,2);
}




/* Comparison operators.
   ----------------------
   
   These are only defined for scalars and strings.
   Comparison operators always return MT_INT, either 0 or 1 

*/


variable_type mac_equals (variable_type x, variable_type y)
{
variable_type v;

	v = nullvar; v.type=MT_INT;
	if ((x.type==MT_INT) && (y.type==MT_INT))
		v.intvalue = x.intvalue == y.intvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
		v.intvalue = x.numvalue == y.numvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
		v.intvalue = x.numvalue == (float)y.intvalue;
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
		v.intvalue = (float)x.intvalue == y.numvalue;
	else if ((x.type==MT_STRING) && (y.type==MT_STRING))
	{
		v.intvalue = strcasecmp (x.strg,y.strg);	/* Note: 0 if equal */
		free (x.strg); free (y.strg);
		v.intvalue = !v.intvalue;
	}
	else
		macro_error (6,current_macro_pos);

	return v;
}
	

variable_type mac_notequals (variable_type x, variable_type y)
{
variable_type v;

	/* Heh. That's an easy one! */
	
	v = mac_equals (x,y);
	v.intvalue = !v.intvalue;
	return v;
}


variable_type mac_gt (variable_type x, variable_type y)
{
variable_type v;

	v = nullvar; v.type=MT_INT;
	if ((x.type==MT_INT) && (y.type==MT_INT))
		v.intvalue = x.intvalue > y.intvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
		v.intvalue = x.numvalue > y.numvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
		v.intvalue = x.numvalue > (float)y.intvalue;
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
		v.intvalue = (float)x.intvalue > y.numvalue;
	else if ((x.type==MT_STRING) && (y.type==MT_STRING))
	{
		if (strcasecmp (x.strg,y.strg) > 0)
			v.intvalue=1;
		else 
			v.intvalue=0;
		free (x.strg); free (y.strg);
	}
	else
		macro_error (6,current_macro_pos);

	return v;
}


variable_type mac_le (variable_type x, variable_type y)
{
variable_type v;

	v = mac_gt (x,y);
	v.intvalue = !v.intvalue;
	return v;
}


variable_type mac_lt (variable_type x, variable_type y)
{
variable_type v;

	v = nullvar; v.type=MT_INT;
	if ((x.type==MT_INT) && (y.type==MT_INT))
		v.intvalue = x.intvalue < y.intvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_REAL))
		v.intvalue = x.numvalue < y.numvalue;
	else if ((x.type==MT_REAL) && (y.type==MT_INT))
		v.intvalue = x.numvalue < (float)y.intvalue;
	else if ((x.type==MT_INT) && (y.type==MT_REAL))
		v.intvalue = (float)x.intvalue < y.numvalue;
	else if ((x.type==MT_STRING) && (y.type==MT_STRING))
	{
		if (strcasecmp (x.strg,y.strg) < 0)
			v.intvalue=1;
		else 
			v.intvalue=0;
		free (x.strg); free (y.strg);
	}
	else
		macro_error (6,current_macro_pos);

	return v;
}


variable_type mac_ge (variable_type x, variable_type y)
{
variable_type v;

	v = mac_lt (x,y);
	v.intvalue = !v.intvalue;
	return v;
}




